// OracleAdoTest.cpp : Defines the entry point for the console application.
//
// References:
// https://msdn.microsoft.com/en-us/library/ms681557(v=vs.85).aspx
// https://msdn.microsoft.com/en-us/library/ms675047%28v=vs.85%29.aspx?f=255&MSPPError=-2147217396
// 

#include "stdafx.h"

#include <iostream>

#include "msado15.tlh"
using namespace std;

//
// Function Declarations
//
inline void TESTHR(HRESULT x) { if FAILED(x) _com_issue_error(x); };
void GetRow();
void PrintComError(_com_error &e);

int main() 
{
	if (FAILED(::CoInitialize(nullptr)))
	 	return -1;

	cout << "C++ ADO using Oracle ODBC project. " << endl;

	GetRow();
	::CoUninitialize();
}

void GetRow() 
{
	// 
	// Use the below SQL to create a single table, single column, and insert one row.  The test
	// is on the ODBC Driver, not the data value.
	// 
	//   CREATE TABLE TABLE1 
    //   (
    //     COLUMN1 TIMESTAMP NOT NULL 
    //   );
    //   INSERT INTO "TABLE1" (COLUMN1) VALUES (TO_TIMESTAMP('2021-06-21 19:06:45.202000000', 'YYYY-MM-DD HH24:MI:SS.FF'));
    // 
	//  strCnn("DRIVER={Oracle in instantclient_19_11};SERVER=SERVER_NAME_HERE;UID=USER_NAME_HERE;PWD=PASSWORD_HERE;DBQ=SERVER_NAME_HERE");
	//
	_bstr_t strCnn("DRIVER={Oracle in instantclient_19_11};SERVER=SERVER_NAME_HERE;UID=USER_NAME_HERE;PWD=PASSWORD_HERE;DBQ=SERVER_NAME_HERE");

	
	//
	// Added to prevent displaying the UID and database password
	//
	wstring wCnx = strCnn;
	wstring wPwd = L"Pwd";// L"Pwd";
	size_t pos = wCnx.find_first_of(wPwd);
	if (pos != wstring::npos)
		wCnx = wCnx.substr(0, pos);
	wcout << L"    The connection string: \"" <<  wCnx.c_str() << L"\"" << endl;
	cout << "    Observe a HEAP CORRUPTION DETECTED dialog, with a DEBUG compliation, before the program exits." << endl << endl;


	try 
	{
		_bstr_t query("SELECT COLUMN1 FROM TABLE1;");
		
		CDatabase db;

		db.OpenEx(strCnn);

		CRecordset rs(&db);
		rs.Open(CRecordset::forwardOnly, query, CRecordset::readOnly);

		CODBCFieldInfo fi;
		rs.GetODBCFieldInfo(short(0), fi);

		CDBVariant val;

		// 
		// SQLGetData Function, 1/19/2017
		// https://docs.microsoft.com/en-us/sql/odbc/reference/syntax/sqlgetdata-function?view=sql-server-ver15
		// When the driver returns fixed-length data, such as an integer or a date structure, the driver ignores 
		// BufferLength and assumes the buffer is large enough to hold the data. It is therefore important for 
		// the application to allocate a large enough buffer for fixed-length data or the driver 
		// will write past the end of the buffer.
		//
		// Passing SQL_C_WCHAR causes buffer overflow / heap corruption too.
		// rs.GetFieldValue(short(0), val, SQL_C_WCHAR); 
		//
		// This form of GetFieldValue uses the type from the database table column to store the data.
		// No corruption occurs since the CDBVariant::m_pdate member contains the value instead of the m_pstringA or m_pstringW.
		// rs.GetFieldValue(short(0), val);
		//
		rs.GetFieldValue(short(0), val, SQL_C_CHAR);

		cout << "Press Enter when ready to exit. ";
		cin.get();
	}
	catch (_com_error &e)
	{
		PrintComError(e);
	}
	catch (std::exception& e)
	{
		cerr << endl << "Received std::exception.  It says: " << e.what() << endl << endl;
	}
	catch (CDBException& err)
	{
		cerr << err.m_strError << endl << endl;
		err.Delete();
	}
	catch (CDBException* err)
	{
		cerr << err->m_strError << endl << endl;
		err->Delete();
	}
	catch (...)
	{
		cerr << endl << "Received unknown exception." << endl << endl;
	}
}

void PrintComError(_com_error &e) 
{
	_bstr_t bstrSource(e.Source());
	_bstr_t bstrDescription(e.Description());

	// Print Com errors.
	cout << "Error:" << endl;
	cout << "\tCode = " << e.Error() << endl;
	cout << "\tCode meaning = " << e.ErrorMessage() << endl;
	cout << "\tSource = " << bstrSource << endl;
	cout << "\tDescription = " << bstrDescription << endl;
}
